package main;

import java.util.ArrayList;
import java.util.Vector;

import structure.State;
import structure.Queue;

/**
 * BoatTripping.class
 * @author vsmital
 * Implementuje zakladni funkcionalitu presunu osob a clunu mezi pobrezimi.
 */
public class BoatTripping {
	
	/**
	 * Stav na levem brehu.
	 */
	private State leftShore = null;

	/**
	 * Stav na pravem brehu.
	 */
	//private State rightShore = null;
	
	/**
	 * Koncovy stav.
	 */
	private final State finalState = new State(0, 0, 0); 
	
	/**
	 * Pomocny stav.
	 */
	private final State aState = new State(-1, -1, -1);
	
	/**
	 * Smer plavby.
	 */
	private int direction;
	
	/**
	 * Konstanta pro plavbu na pravy breh.
	 */
	private final int LEFT_DIR = 0;
	
	/**
	 * Konstanta pro plavbu na levy breh.
	 */
	private final int RIGHT_DIR = 1;
	
	/**
	 * Struktura vyhledavaciho stromu.
	 */
	private Queue open = null;
	private Queue closed = null;
	
	/**
	 * Struktura pro nalezeni cesty.
	 */
	private ArrayList pathFinder = null;
	
	/**
	 * Struktura prechodu.
	 */

	/**
	 * Konstruktor.
	 *
	 */
	public BoatTripping() {
		leftShore = new State(3, 3, 1);
		open = new Queue();
		closed = new Queue();
		pathFinder = new ArrayList();
	}
	
	/**
	 * Metoda vyhledavajici posloupnost kroku vedoucich ke koncovemu stavu na levem brehu.
	 *
	 */
	public void process() {
		if(leftShore.equals(finalState))
			return;
		open.addNode(leftShore);
		do{
			State state = open.removeNode(0);
			Vector vector = new Vector(6);
			vector.add(0, state);
			System.out.print("[" + state.getMisNum() + ", " + state.getCanNum() + ", " + state.getBoatNum() + "]");
			System.out.print(" -------- ");
			if(state.equals(finalState)) {
				System.out.println("SUCCESS");
				break;
			}
			closed.addNode(state);
			int m1 = state.getMisNum();
			int c1 = state.getCanNum();
			int b1 = state.getBoatNum();
			if(b1 == 1)
				direction = RIGHT_DIR;
			else 
				direction = LEFT_DIR;
			State lShore = new State(m1, c1, b1);
			State rShore = new State(0, 0, 0);
			int m = lShore.getMisNum();
			m = 3 - m;
			int c = lShore.getCanNum();
			c = 3 - c;
			int b = lShore.getBoatNum();
			b = 1 - b;
			rShore.addMiss(m);
			rShore.addCan(c);
			rShore.addBoat(b);
			State resultState = oneMissionareOut(direction, lShore, rShore);
			if(resultState != null)
				if(!closed.contains(resultState) && !open.contains(resultState)) {
					open.addNode(resultState);
					vector.add(1, resultState);
				}
				else
					vector.add(1, aState);
			else
				vector.add(1, aState);
			
			resultState = twoMissionaresOut(direction, lShore, rShore);
			if(resultState != null)
				if(!closed.contains(resultState) && !open.contains(resultState)) {
					open.addNode(resultState);
					vector.add(2, resultState);
				}
				else
					vector.add(2, aState);
			else
				vector.add(2, aState);
			
			resultState = oneCanibalOut(direction, lShore, rShore);
			if(resultState != null)
				if(!closed.contains(resultState) && !open.contains(resultState)) {
					open.addNode(resultState);
					vector.add(3, resultState);
				}
				else
					vector.add(3, aState);
			else
				vector.add(3, aState);
			
			resultState = twoCanibalsOut(direction, lShore, rShore);
			if(resultState != null)
				if(!closed.contains(resultState) && !open.contains(resultState)) { 
					open.addNode(resultState);
					vector.add(4, resultState);
				}
				else
					vector.add(4, aState);
			else
				vector.add(4, aState);
			
			resultState = misAndCanOut(direction, lShore, rShore);
			if(resultState != null)
				if(!closed.contains(resultState) && !open.contains(resultState)) {
					open.addNode(resultState);
					vector.add(5, resultState);
				}
				else
					vector.add(5, aState);
			else
				vector.add(5, aState);
			
			for(int i = 1; i < 6; i++) {
				State s = (State) vector.get(i);
				System.out.print(s.toString() + ", ");
			}
			System.out.println();
			pathFinder.add(vector);
		}while(true);
		open = null;
		closed = null;
		pathLookup(pathFinder);
	}
	
	/**
	 * Metoda, ktera modeluje situaci plavby jednoho misionare pres reku.
	 * @param dir - smer plavby.
	 * @param left - stav na levem brehu.
	 * @param right - stav na pravem brehu.
	 * @return - vraci State, pokud je mozne tuto operaci provest, jinak null.
	 */
	public State oneMissionareOut(int dir, State left, State right) {
		int lm = left.getMisNum();
		int lc = left.getCanNum();
		int lb = left.getBoatNum();
		
		if(dir == LEFT_DIR && lb > 0)
			return (State) null;
		
		int rm = right.getMisNum();
		int rc = right.getCanNum();
		int rb = right.getBoatNum();
		
		if(dir == RIGHT_DIR && rb > 0)
			return (State) null;
		
		boolean status = missionareOutControl(lm, lc, rm, rc, 1, dir);
		State result = new State(lm, lc, lb);
		if(status) {
			if(dir == LEFT_DIR) {
				result.addMiss(1);
				result.addBoat(1);
			}
			else{
				result.removeMiss(1);
				result.remBoat(1);
			}
		}
		if(status)
			return result;
		else
			return (State) null;
	}
	
	/**
	 * Metoda, ktera modeluje situaci plavby dvou misionaru pres reku.
	 * @param dir - smer plavby
	 * @param left - levy breh.
	 * @param right - pravy breh.
	 * @return - vraci State, pokud je mozne tuto operaci provest, jinak null.
	 */
	public State twoMissionaresOut(int dir, State left, State right) {
		int lm = left.getMisNum();
		int lc = left.getCanNum();
		int lb = left.getBoatNum();
		
		if(dir == LEFT_DIR && lb > 0)
			return (State) null;
		
		int rm = right.getMisNum();
		int rc = right.getCanNum();
		int rb = right.getBoatNum();
		
		if(dir == RIGHT_DIR && rb > 0)
			return (State) null;
		
		boolean status = missionareOutControl(lm, lc, rm, rc, 2, dir);
		State result = new State(lm, lc, lb);
		if(status) {
			if(dir == LEFT_DIR) {
				result.addMiss(2);
				result.addBoat(1);
			}
			else{
				result.removeMiss(2);
				result.remBoat(1);
			}
		}
		if(status)
			return result;
		else
			return (State) null;
	}
	
	/**
	 * Metoda pro kontrolu zakladni premisy - teda ze pocet misionaru na brehu neni mensi nez pocet kanibalu. 
	 * @param lm - pocet misionaru na levem brehu.
	 * @param lc - pocet kanibalu na levem brehu.
	 * @param rm - pocet misionaru na pravem brehu.
	 * @param rc - pocet kanibalu na pravem brehu.
	 * @param num - pocet presunovanych misionaru.
	 * @param dir - smer plavby.
	 * @return - vraci true, pokud je mozne operaci provest, jinak false.
	 */
	private boolean missionareOutControl(int lm, int lc, int rm, int rc, int num, int dir) {
		boolean status = false;
		if(dir == LEFT_DIR) {
			if((((lm + num) >= lc)) && ((lm + num) <= 3) && ((rm - num) >= 0)) {
				if(((rm - num > 0) && (rm - num >= rc)) || ((rm - num) == 0))
					status = true;
			}
		}
		else{
			if(((rm + num) >= rc) && ((rm + num) <= 3) && ((lm - num) >= 0)) {
				if(((lm - num > 0) && (lm - num >= lc)) || ((lm - num) == 0))
					status = true;
			}
		}
		return status;
	}
	
	/**
	 * Metoda, ktera modeluje situaci plavby jednoho kanibala pres reku.
	 * @param dir - smer plavby.
	 * @param left - levy breh.
	 * @param right - pravy breh.
	 * @return - vraci State, pokud je mozne tuto operaci provest, jinak null.
	 */
	public State oneCanibalOut(int dir, State left, State right) {
		int lm = left.getMisNum();
		int lc = left.getCanNum();
		int lb = left.getBoatNum();
		
		if(dir == LEFT_DIR && lb > 0)
			return (State) null;
		
		int rm = right.getMisNum();
		int rc = right.getCanNum();
		int rb = right.getCanNum();
		
		if(dir == RIGHT_DIR && rb > 0)
			return (State) null;
		
		boolean status = canibalOutControl(lm, lc, rm, rc, 1, dir);
		State result = new State(lm, lc, lb);
		if(status) {
			if(dir == LEFT_DIR) {
				result.addCan(1);
				result.addBoat(1);
			}
			else{
				result.removeCan(1);
				result.remBoat(1);
			}
		}
		if(status)
			return result;
		else
			return (State) null;
	}
	
	/**
	 * Metoda, ktera modeluje situaci plavby dvou kanibalu pres reku.
	 * @param dir - smer plavby
	 * @param left - levy breh.
	 * @param right - pravy breh.
	 * @return - vraci State, pokud je mozne tuto operaci provest, jinak null.
	 */
	public State twoCanibalsOut(int dir, State left, State right) {
		int lm = left.getMisNum();
		int lc = left.getCanNum();
		int lb = left.getBoatNum();
		
		if(dir == LEFT_DIR && lb > 0)
			return (State) null;
		
		int rm = right.getMisNum();
		int rc = right.getCanNum();
		int rb = right.getBoatNum();
		
		if(dir == RIGHT_DIR && rb > 0)
			return (State) null;
		
		boolean status = canibalOutControl(lm, lc, rm, rc, 2, dir);
		State result = new State(lm, lc, lb);
		if(status) {
			if(dir == LEFT_DIR) {
				result.addCan(2);
				result.addBoat(1);
			}
			else{
				result.removeCan(2);
				result.remBoat(1);
			}
		}
		if(status)
			return result;
		else
			return (State) null;
	}
	
	/**
	 * Pomocna metoda
	 * @param lm - pocet misionaru na levem brehu.
	 * @param lc - pocet kanibalu na levem brehu.
	 * @param rm - pocet misionaru na pravem brehu.
	 * @param rc - pocet kanibalu na pravem brehu.
	 * @param num - pocet prevazenych kanibalu.
	 * @param dir - smer plavby.
	 * @return vraci true, pokud lze operaci provest, jinak false.
	 */
	private boolean canibalOutControl(int lm, int lc, int rm, int rc, int num, int dir) {
		boolean status = false;
		if(dir == LEFT_DIR) {
			if((((lc + num) <= lm) || (lm == 0)) && ((lc + num) <= 3) && ((rc - num) >= 0)) { 
				if((((rc - num) >= 0) && ((rc - num) <= rm)) || (rm == 0))
					status = true;
			}
		}
		else{
			if((((rc + num) <= rm) || (rm == 0)) && ((rc + num) <= 3) && ((lc - num) >= 0)) {
				if(((lc - num) >= 0) && ((lc - num) <= lm) || (lm == 0))
					status = true;
			}
		}
		return status;
	}
	
	/**
	 * Modeluje pripad kdy lodi jedou spolecne misionar a kanibal
	 * @param dir - smer plavby.
	 * @param left - levy breh.
	 * @param right - pravy breh. 
	 * @return
	 */
	public State misAndCanOut(int dir, State left, State right) {
		int lm = left.getMisNum();
		int lc = left.getCanNum();
		int lb = left.getBoatNum();
		
		if(dir == LEFT_DIR && lb > 0)
			return (State) null;
		
		int rm = right.getMisNum();
		int rc = right.getCanNum();
		int rb = right.getBoatNum();
		
		if(dir == RIGHT_DIR && rb > 0)
			return (State) null;
		
		boolean status = canAndMisOutControl(lm, lc, rm, rc, dir);
		State result = new State(lm, lc, lb);
		if(status) {
			if(dir == LEFT_DIR) {
				result.addCan(1);
				result.addMiss(1);
				result.addBoat(1);
			}
			else{
				result.removeCan(1);
				result.removeMiss(1);
				result.remBoat(1);
			}
		}
		if(status)
			return result;
		else
			return (State) null;
	}
	
	
	/**
	 * Pomocna metoda.
	 * @param lm - pocet misionaru na levem brehu.
	 * @param lc - pocet kanibalu na levem brehu.
	 * @param rm - pocet misionaru na pravem brehu.
	 * @param rc - pocet kanibalu na pravem brehu.
	 * @param dir - smer plavby.
	 * @return - vraci true, pokud je operaci mozne provest, jinak false. 
	 */
	private boolean canAndMisOutControl(int lm, int lc, int rm, int rc, int dir) {
		boolean status = false;
		if(dir == LEFT_DIR) {
			if(((lm + 1) >= (lc + 1)) && ((lm + 1) <= 3) && ((lc + 1) <= 3) && ((rm - 1) >= 0) && ((rc - 1) >= 0)) {
				if(((rm - 1) >= (rc - 1)) || ((rm - 1) == 0))
					status = true;
			}
		}
		else{
			if(((rm + 1) >= (rc + 1)) && ((rm + 1) <= 3) && ((rc + 1) <= 3) && ((lm - 1) >= 0) && ((lc - 1) >=0)) {
				if(((lm - 1) >= (rc - 1)) || ((lm - 1) == 0))
						status = true;
			}
		}
		return status;
	}
	
	/**
	 * Metoda main
	 * @param args - vstupni argumenty z prikazove radky - nejsou vyuzity.
	 */
	public static void main(String[] args) {
		BoatTripping trip = new BoatTripping();
		trip.process();
	}
	
	private void pathLookup(ArrayList list) {
		ArrayList result = new ArrayList();
		System.out.println();
		System.out.println();
		System.out.println();
		System.out.println("PATH: ");
		int listSize = list.size();
		int size = listSize - 1;
		State state = null;
		State searchedState = finalState;
		int j = 0;
		do{
			Vector vector = (Vector) list.get(size);
			for(int i = 1; i < 6; i++) {
				state = (State) vector.get(i);
				if(state.equals(searchedState)){
					searchedState = (State) vector.get(0);
					result.add(j, searchedState);
					j++;
				}
				if(searchedState.equals(leftShore))
					break;
			}
			size--;
			if(size == 0)
				break;
		}while(true);
		
		//System.out.print("  ");
		System.out.print("[" + leftShore.getMisNum() + ", " +
				leftShore.getCanNum() + ", " + leftShore.getBoatNum() + "]");
		System.out.print(",");
		for(int i = result.size() - 1; i >= 0; i--) {
			state = (State) result.get(i);
			System.out.print("[" + state.getMisNum() + ", " +
					state.getCanNum() + ", " + state.getBoatNum() + "]");
			if(i > 0)
				System.out.print(",");
		}
		System.out.print(",");
		System.out.print("[" + finalState.getMisNum() + ", " +
				finalState.getCanNum() + ", " + finalState.getBoatNum() + "]");
		
	}
}
