package structure;


/**
 * State.class
 * @author vsmital
 * uchovava informace o poctu misionaru, kanibalu a clunu a urcitem brehu.
 */
public class State {

	/**
	 * Pocet misionaru na brehu.
	 */
	private int missionares = 0;
	/**
	 * Pocet kanibalu na brehu.
	 */
	private int cannibals = 0;
	/**
	 * Pocet clunu na brehu.
	 */
	private int boat = 0;
	
	/**
	 * Konstruktor tridy State
	 * @param m - iniciani pocet misionaru.
	 * @param c - inicialni pocet kanibalu.
	 * @param b - inicialni pocet clunu.
	 */
	public State(int m, int c, int b) {
		missionares = m;
		cannibals = c;
		boat = b;
	}
	
	/**
	 * Vraci pocet misionaru
	 * @return
	 */
	public int getMisNum() {
		return missionares;
	}
	
	/**
	 * Vraci pocet kanibalu.
	 * @return
	 */
	public int getCanNum() {
		return cannibals;
	}
	
	/**
	 * Vraci pocet clunu.
	 * @return
	 */
	public int getBoatNum() {
		return boat;
	}
	
	/**
	 * Prida misionare.
	 * @param num - pocet pridavanych misionaru
	 * @return - vraci true pokud lze operaci pridani provest, jinak false.
	 */
	public boolean addMiss(int num) {
		if((missionares + num) <= 3) {
			missionares = missionares + num;
			return true;
		}
		else
			return false;
	}
	
	/**
	 * Odebere misionare.
	 * @param num - pocet odebiranych misionaru.
	 * @return - vraci true pokud lze operace odebirani provest, jinak false.
	 */
	public boolean removeMiss(int num) {
		if((missionares - num) >= 0) {
			missionares = missionares - num;
			return true;
		}
		else
			return false;
	}
	
	/**
	 * Prida kanibaly.
	 * @param num - pocet pridavanych kanibalu.
	 * @return - vraci true pokud lze operace pridani provest, jinak false.
	 */
	public boolean addCan(int num) {
		if((cannibals + num) <= 3) {
			cannibals = cannibals + num;
			return true;
		}
		else
			return false;
	}
	
	/**
	 * Odebira kanibaly.
	 * @param num - pocet odebiranych kanibalu.
	 * @return - vraci true pokud lze operaci odebirani provest, jinak false.
	 */
	public boolean removeCan(int num) {
		if((cannibals - num) >= 0) {
			cannibals = cannibals - num;
			return true;
		}
		else
			return false;
	}
	
	/**
	 * Pridava clun.
	 * @param num - pocet pridavanych clunu - vzdy num = 1.
	 * @return - vraci true pokud lze operaci provest, jinak false.
	 */
	public boolean addBoat(int num) {
		boat = boat + num;
		return true;
	}
	
	/**
	 * Odebira clun.
	 * @param num - pocet odebiranych clunu - vzdy num = 1;
	 * @return
	 */
	public boolean remBoat(int num) {
		boat = boat - num;
		return true;
	}
	
	/**
	 * Metoda pro porovnavani Objektu tridy State
	 * @param state - object tridy State k porovnavani
	 * @return
	 */
	public boolean equals(State state) {
		if(this.missionares == state.missionares && this.cannibals == state.cannibals
				&& this.boat == state.boat)
			return true;
		else
			return false;
	}
	
	public String toString() {
		String result = null;
		result = "[" + missionares + ", " + cannibals + ", " + boat + "]";
		return result;
	}
}
